function BeneCast:OnInitialize()
	self:FillSpellInfo();
end

function BeneCast:OnEnable()

end

function BeneCast:OnDisable()

end

function BeneCast:GetProfile(pf, unit)
	if not BeneCastProfiles then
		return;
	end
	
	if not BeneCastProfiles[profileid] then
		self:ResetProfile(profileid);
	end
	
	if not BeneCastProfiles[profileid].SnapTo then
		BeneCastProfiles[profileid].SnapTo = {};
	end
	
	if unit then
		if not BeneCastProfiles[profileid].SnapTo[unit] then
			BeneCastProfiles[profileid].SnapTo[unit] = {
				Y = 0,
				X = 0,
				Point = 'TOPLEFT',
				RelativePoint = 'TOPLEFT',
				Orientation = BeneCastProfiles[profileid].Orientation,
			};
		end
		return BeneCastProfiles[profileid], BeneCastProfiles[profileid].SnapTo[unit];
	else
		return BeneCastProfiles[profileid];
	end
end

function BeneCast:GetCurrentProfileID()
	if wv.ConsiderInCombat then
		return 'combat';
	else
		return 'ooc';
	end
end

-- Function to resize the buttons of a profile
function BeneCast:ResizeButtonsForProfile(pf)

	if BeneCast.Panels[pf] == nil then
		BeneCast.Panels[pf] = {};
	end
	
	-- Resize the buttons of every panel
	for unit, panel in pairs (BeneCast.Panels[pf]) do
		panel:ResizeButtons(i);
	end
	
end

-- Resets data tables
function BeneCast:ResetData()

	-- Clear the BC_spell_data table
	_ClearTable(wv.SpellData);
	
	-- Reload the spell data
	self:LoadSpellData();
	
end

function BeneCast:Debug(msg)
	if ( BeneCastSetup.Debug and DEFAULT_CHAT_FRAME ) then
		self:Print('Debug: ' .. msg);
	end
end

-- *****************************************************************************
-- BeneCast Spell Loading function
-- *****************************************************************************

-- Fill in the wv.Spells table by iterating over spells in spellbook
function BeneCast:LoadSpellData()
	self:Debug('Parsing Spell Data from spellbook');
	
	wv.Spells = _ClearTable(wv.Spells);
	
	-- Iterator to move through the spellbook
	local i = 1;
	local spell_name, spell_rank, rankno;
	local spell, oldspell, spellinfo, rootid;
	local spellgroup;

	while true do
		
		-- Spell at position i in the player's spellbook
		spell_name, spell_rank = GetSpellName(i, SpellBookFrame.bookType);
		
		-- If there is no spell at i, break out of the loop
		if not spell_name then
			do break end
		end
		
		_, _, rankno = string.find(spell_rank,'%d+'); -- Find any string of numbers in the rank
		if rankno then
			rankno = tonumber(rankno);
		end
			
		spellinfo = wv.SpellInfo[spell_name];
		if spellinfo then
			spell = {
				spell = true,
				name = spell_name,
				rank = spell_rank,
				rankno = rankno,
				spellbookid = i,
				infoid = spellinfo.id,
				group = spellinfo.group,
				groupid = 0,
				type = spellinfo.type,
				target = spellinfo.target,
				stances = _CopyTable(spellinfo.stances,false),
				code = spellinfo.code,
				cures = _CopyTable(spellinfo.cures,false),
				overrides = _CopyTable(spellinfo.overrides,false),
				icon = spellinfo.icon,
				minrange = spellinfo.minrange,
				maxrange = spellinfo.maxrange,
				sorting = spellinfo.sorting,
				uninteractible = spellinfo.uninteractible,
			};
			if spell.sorting == nil then
				spell.sorting = 1;
			end
			if BeneCast.GroupInfo[spell.group] then
				for grpid, spellid in pairs(BeneCast.GroupInfo[spell.group]) do
					if spellid == spell.infoid then
						spell.groupid = grpid;
					end
				end
			end
			oldspell = wv.Spells[spell.group];
			if oldspell then	
				if (spell.groupid > oldspell.groupid) or (spell.groupid == oldspell.groupid and spell.rankno > oldspell.rankno) then
					wv.Spells[spell.group] = spell;
				end
			else
				wv.Spells[spell.group] = spell;
			end
						
		end

		i = i + 1;

	end
	
end

function BeneCast:AddButtonThatFlashes(button)
	if not mv.ButtonsThatFlash[pf] then
		mv.ButtonsThatFlash[pf] = {};
	end
	if not mv.ButtonsThatFlash[pf][unit] then
		mv.ButtonsThatFlash[pf][unit] = {};
	end
	table.add(mv.ButtonsThatFlash[pf][unit],button);
end

-- *****************************************************************************
-- BeneCastOptionFrameTab functions
-- *****************************************************************************

function BeneCastOptionFrameTab_OnClick()

	-- Show the correct subframe based on which tab was clicked
	-- Hide all subframes that do not have matching ID numbers, show the matching subframe
	for subframe, subframename in pairs (BeneCastOptionFrameSubframes) do
		if this:GetID() == getglobal(subframename):GetID() then
			getglobal(subframename):Show();
		else
			getglobal(subframename):Hide();
		end
	end
	
	-- Play a sound for player feedback
	PlaySound('igCharacterInfoTab');
	
	-- Set the glowing checked status for the correct tab
	for x in pairs (BeneCastOptionFrameSubframes) do
		if ( x == this:GetID() ) then
			this:Disable()
			this:SetChecked(1);
		else
			getglobal('BeneCastOptionFrameTab' .. x):Enable();
			getglobal('BeneCastOptionFrameTab' .. x):SetChecked(nil);
		end
	end
	
end

-- *****************************************************************************
-- BeneCastOptionsCheckButton functions
-- *****************************************************************************

function BeneCastOptionsCheckButton_OnClick()

	local dontredraw = nil;
	if BeneCastSetup.currentoptionid == nil then
		BeneCastSetup.currentoptionid = 1;
	end
	local pf = BeneCastSetup.currentoptionid;
  if BeneCastProfiles[pf] == nil then
    BeneCastProfiles[pf] = {};
  end
  
	-- Set the BeneCastProfiles for the option to take on the value of the checkbutton
	for index, value in pairs (BeneCastOptionFrameCheckButtons) do
		if ( value.index == this:GetID() ) then
			BeneCastProfiles[pf][value.cvar] = this:GetChecked();
			-- If this is the tooltip option check to see if it's enabled
			-- If not disable the tooltip name option checkbutton
			if ( value.index == 5 ) then
				if ( this:GetChecked() ) then
					getglobal('BeneCastOptionCheckButton6'):Enable();
				else
					getglobal('BeneCastOptionCheckButton6'):Disable();
				end
			-- If this is the notification option check to see if it's enabled
			-- If not disable all sub notification options
			elseif ( value.index == 8 ) then
				if ( this:GetChecked() ) then
					getglobal('BeneCastOptionCheckButton9'):Enable();
					getglobal('BeneCastOptionCheckButton10'):Enable();
					getglobal('BeneCastOptionCheckButton11'):Enable();
					getglobal('BeneCastOptionCheckButton13'):Enable();
					getglobal('BeneCastOptionCheckButton14'):Enable();
					getglobal('BeneCastOptionCheckButton15'):Enable();
					getglobal('BeneCastOptionCheckButton16'):Enable();
					getglobal('BeneCastOptionCheckButton17'):Enable();
					getglobal('BeneCastOptionCheckButton18'):Enable();
					getglobal('BeneCastOptionCheckButton19'):Enable();
					getglobal('BeneCastOptionCheckButton21'):Enable();
				else
					getglobal('BeneCastOptionCheckButton9'):Disable();
					getglobal('BeneCastOptionCheckButton10'):Disable();
					getglobal('BeneCastOptionCheckButton11'):Disable();
					getglobal('BeneCastOptionCheckButton13'):Disable();
					getglobal('BeneCastOptionCheckButton14'):Disable();
					getglobal('BeneCastOptionCheckButton15'):Disable();
					getglobal('BeneCastOptionCheckButton16'):Disable();
					getglobal('BeneCastOptionCheckButton17'):Disable();
					getglobal('BeneCastOptionCheckButton18'):Disable();
					getglobal('BeneCastOptionCheckButton19'):Disable();
					getglobal('BeneCastOptionCheckButton21'):Disable();
				end
			-- If this is the heal notification option check to see if it's enabled
			-- If not disable all heal max notification
			elseif ( value.index == 16 ) then
				if ( this:GetChecked() ) then
					getglobal('BeneCastOptionCheckButton17'):Enable();
				else
					getglobal('BeneCastOptionCheckButton17'):Disable();
				end
			-- If this is the hide minimap button option check to see if it's enabled
			-- If so then hide the minimap button
			elseif ( value.index == 4 ) then
				if ( this:GetChecked() ) then
					getglobal('BeneCastMinimapButton'):Hide();
				else
					getglobal('BeneCastMinimapButton'):Show();
				end	
			--Allow Raidpets to be clicked only if Raidmembers is checked
			elseif ( value.index == 24 ) then
				local cbutton = getglobal('BeneCastOptionCheckButton25');
				if ( this:GetChecked() ) then
					cbutton:Enable();
				else
					BeneCastProfiles[pf].ShowRaidPets = false;
					cbutton:SetChecked(BeneCastProfiles[pf].ShowRaidPets);
					cbutton:Disable();
				end
				BC_UpdateRaid = true;
				dontredraw = true;
			--Redraw raidbuttons to show/hide raidpets
			elseif ( value.index == 25 ) then
				BC_UpdateRaid = true;
				dontredraw = true;
			--Redraw raidbuttons to overlap or not to overlap
			elseif ( value.index == 1 ) then
				BC_UpdateRaid = true;
				dontredraw = true;
			end
			do break end; -- Break for efficiency
		end
	end
	-- Play a sound for player feedback
	if ( this:GetChecked() ) then
		PlaySound('igMainMenuOptionCheckBoxOff');
	else
		PlaySound('igMainMenuOptionCheckBoxOn');
	end
	
	-- Redraw everyone's buttons
	if not dontredraw then
		if pf == BeneCastSetup.currentid then
			for i = 1, 5 do
				BeneCast_UpdateButtons(i);
			end
			for i = 46, 50 do
				BeneCast_UpdateButtons(i);
			end
			for i = 91, 93 do
				BeneCast_UpdateButtons(i);
			end
		end
		if pf == BeneCastSetup.currentraidid then
			for i = 6, 45 do
				BeneCast_UpdateButtons(i);
			end
			for i = 51, 90 do
				BeneCast_UpdateButtons(i);
			end
		end
	end

end

-- *****************************************************************************
-- BeneCastProfileCheckButton functions
-- *****************************************************************************

function BeneCastProfileCheckButton_OnClick()

	if not this:GetChecked() then
		this:SetChecked(true);
	else
		local profilebox = getglobal('BeneCastProfileName' .. this.var2);
		if profilebox:GetText() then
			local found;
			for k, v in pairs(BeneCastProfiles) do
				if v.profilename == profilebox:GetText() then
					found = k;
					do break end;
				end
			end
			if found then
				if this.var1 == 'profilecombat' then
					for i = 1, 5 do
						if i ~= this.var2 then
							getglobal('BeneCastCombatCheck' .. i):SetChecked(false);
						end
					end
					BeneCastSetup['PrefferedInCombatProfile'] = found;
				elseif this.var1 == 'profileoutofcombat' then
					for i = 1, 5 do
						if i ~= this.var2 then
							getglobal('BeneCastOutOfCombatCheck' .. i):SetChecked(false);
						end
					end
					BeneCastSetup['PrefferedOutOfCombatProfile'] = found;
				elseif this.var1 == 'profilecombatraid' then
					for i = 1, 5 do
						if i ~= this.var2 then
							getglobal('BeneCastCombatCheck' .. i .. 'r'):SetChecked(false);
						end
					end
					BeneCastSetup['PrefferedCombatProfileRaid'] = found;
				elseif this.var1 == 'profilecoutofombatraid' then
					for i = 1, 5 do
						if i ~= this.var2 then
							getglobal('BeneCastOutOfCombatCheck' .. i .. 'r'):SetChecked(false);
						end
					end
					BeneCastSetup['PrefferedOutOfCombatProfileRaid'] = found;
				end
				BeneCast_UpdateCurrentProfile();
			else
				PutMsg('No profile found with name=' .. profilebox:GetText());
				this:SetChecked(false);
			end
		else
			this:SetChecked(false);
		end
	end
				
end

function BeneCast_UpdateProfileCheckButtons()

	local profindex = 1;
	for k, v in pairs(BeneCastProfiles) do
		local profilebox = getglobal('BeneCastProfileName' .. profindex);
		profilebox:SetText(v.profilename);
		profilebox.oldtext = v.profilename;
		profindex = profindex + 1;
	end
	for i = profindex, 5 do
		local profilebox = getglobal('BeneCastProfileName' .. i);
		profilebox:SetText('');
		profilebox.oldtext = '';
	end

	for i = 1, 5 do
		local profilebox = getglobal('BeneCastProfileName' .. i);
		local checkbutton;
		checkbutton = getglobal('BeneCastCombatCheck' .. i);
		checkbutton:SetChecked(profilebox:GetText() == BeneCastSetup.PrefferedInCombatProfile);
		checkbutton = getglobal('BeneCastOutOfCombatCheck' .. i);
		checkbutton:SetChecked(profilebox:GetText() == BeneCastSetup.PrefferedOutOfCombatProfile);
		checkbutton = getglobal('BeneCastCombatCheck' .. i .. 'r');
		checkbutton:SetChecked(profilebox:GetText() == BeneCastSetup.PrefferedInCombatProfileRaid);
		checkbutton = getglobal('BeneCastOutOfCombatCheck' .. i .. 'r');
		checkbutton:SetChecked(profilebox:GetText() == BeneCastSetup.PrefferedOutOfCombatProfileRaid);
	end
end

function BeneCast_UpdateCurrentProfile()

	if BC_ConsiderInCombat then
		BeneCastSetup.currentid = BeneCastSetup.PrefferedInCombatProfile;
		BeneCastSetup.currentraidid = BeneCastSetup.PrefferedInCombatProfileRaid;
	else
		BeneCastSetup.currentid = BeneCastSetup.PrefferedOutOfCombatProfile;
		BeneCastSetup.currentraidid = BeneCastSetup.PrefferedOutOfCombatProfileRaid;
	end
	
	if BeneCastProfiles[BeneCastSetup.currentid] == nil then
	  BeneCastProfiles[BeneCastSetup.currentid] = {};
	end
	if BeneCastProfiles[BeneCastSetup.currentraidid] == nil then
	  BeneCastProfiles[BeneCastSetup.currentraidid] = {};
	end
	
	if BeneCastProfiles[BeneCastSetup.currentid] then
		for i = 1, 5 do
			BeneCast_UpdateButtons(i);
		end
		for i = 46, 50 do
			BeneCast_UpdateButtons(i);
		end
		for i = 91, 93 do
			BeneCast_UpdateButtons(i);
		end
	end
	if BeneCastProfiles[BeneCastSetup.currentraidid] then
		for i = 6, 45 do
			BeneCast_UpdateButtons(i);
		end
		for i = 51, 90 do
			BeneCast_UpdateButtons(i);
		end
	end
end

-- *****************************************************************************
-- BeneCastSpellCheckButton functions
-- *****************************************************************************

function BeneCastSpellCheckButton_OnClick()

	if BeneCastSetup.currentoptionid == nil then
		BeneCastSetup.currentoptionid = 1;
	end
	local profileid = BeneCastSetup.currentoptionid;
	if BeneCastProfiles[profileid] == nil then
	  BeneCastProfiles[profileid] = {};
	end
	for i = 1, 10 do
		if BeneCastProfiles[profileid][i] == nil then
			BeneCastProfiles[profileid][i] = {};
		end
	end

	-- Put some things in a local in the beginning for efficiency (thnx Shimavak)
	local spellname = getglobal( this:GetName() .. 'Text' ):GetText();
	local parentid = this:GetParent():GetID();
	
	-- Set the BeneCastProfiles for the option to take on the value of the checkbutton
	local spelltype;
	if ( spellname == BENECAST_STRINGS.TEXT_ASSIST ) then
		spelltype = 'assist';
	else
		spelltype = BeneCast_SpellTypes[BC_class][spellname];
	end
	BeneCastProfiles[profileid][parentid][spelltype] = this:GetChecked();
	if ( parentid == 1 and BeneCastProfiles[profileid]['PlayerAsDefault'] ) then
		for i = 2, 10 do
			BeneCastProfiles[profileid][i][spelltype] = this:GetChecked();
			getglobal(BeneCastOptionFrameSubframes[i] .. 'Button' .. this:GetID()):SetChecked(this:GetChecked());
		end
	end
	-- Play a sound for player feedback
	if ( this:GetChecked() ) then
		PlaySound('igMainMenuOptionCheckBoxOff');
	else
		PlaySound('igMainMenuOptionCheckBoxOn');
	end
	
	-- Redraw everyone's buttons
	for i in pairs (BC_targets) do
		BeneCast_UpdateButtons(i);
	end

end

-- *****************************************************************************
-- BeneCast_Tooltip functions
-- *****************************************************************************

function BeneCast_Tooltip_OnLoad()

	-- For some unknown reason, when doing reloadui, it will not parse the OnLoad from
	-- 	the XML file for BeneCast_Tooltip...so we're doing that here.
	this:SetOwner(WorldFrame, "ANCHOR_NONE");
end

-- *****************************************************************************
-- BeneCastPanelManager functions
-- *****************************************************************************

function BeneCastPanelManager_OnLoad()

	-- Set BC_class to the player's class
	BC_class = UnitClass('player');
	
	-- Register for Events
	this:RegisterEvent('ADDON_LOADED');
	this:RegisterEvent('PLAYER_ENTERING_WORLD');
	this:RegisterEvent('PLAYER_LEAVING_WORLD');
	this:RegisterEvent('VARIABLES_LOADED');
	this:RegisterEvent('SPELLS_CHANGED');
	this:RegisterEvent('SPELLCAST_FAILED');
	this:RegisterEvent('SPELLCAST_INTERRUPTED');
	this:RegisterEvent('SPELLCAST_STOP');
	this:RegisterEvent('SPELLCAST_START');
	this:RegisterEvent('UNIT_SPELLCAST_CHANNEL_START');
	this:RegisterEvent('UNIT_SPELLCAST_CHANNEL_STOP');
	this:RegisterEvent('PARTY_MEMBERS_CHANGED');
	this:RegisterEvent('PLAYER_TARGET_CHANGED');
	this:RegisterEvent('PLAYER_FOCUS_CHANGED');
	this:RegisterEvent('UNIT_AURA');
	this:RegisterEvent('PLAYER_AURAS_CHANGED');
	this:RegisterEvent('UNIT_AURASTATE');
	this:RegisterEvent('UPDATE_SHAPESHIFT_FORMS');
	this:RegisterEvent('PLAYER_ALIVE');
	this:RegisterEvent('PLAYER_DEAD');
	this:RegisterEvent('PLAYER_UNGHOST');
	this:RegisterEvent('RAID_ROSTER_UPDATE');
	this:RegisterEvent('UNIT_HEALTH');
	this:RegisterEvent('UNIT_PET');
	this:RegisterEvent('UNIT_TARGET');
	this:RegisterEvent('PLAYER_REGEN_DISABLED');
	this:RegisterEvent('PLAYER_REGEN_ENABLED');

end

function BeneCastPanelManager_OnEvent()

	if ( event == 'ADDON_LOADED' and arg1 == 'BeneCast' ) then
		BeneCast.db = LibStub("AceDB-3.0"):New("MyAddonDB");
		LibStub("AceConfig-3.0"):RegisterOptionsTable("BeneCast", options, {"benecast", "bc"});
			
		-- Reset the data if there BeneCast is severly outdated
		if ( not BeneCastSetup or not BeneCastProfiles or
		     table.getn(BeneCastProfiles) == 0 or
		     not BeneCastSetup.Version or BeneCastSetup.Version < BENECAST_VERSION ) then
			PutMsg('Resetting settings')
			BeneCast_ResetProfiles();
		end
		BeneCast_UpdateCurrentProfile();
		
		-- Populating all of our lists of constants now so we don't have to do it repeatedly.
		for i in pairs (BC_targets) do
			BC_attached_targets[i] = false;
		end
		
		for i,buffInfo in pairs (BC_buffs_in_effect) do
			BC_buffs_in_effect_by_name[buffInfo.name] = i;
		end
		
		-- Load up spell data once AddOn has been loaded
		BeneCast_ResetData();
		
		--Code from BeneCastOptionFrame
		BeneCast_LoadOptions();
		BeneCast_UpdateProfileCheckButtons();
		BeneCast_ResizeButtons();
		BeneCast_ClearRaid();
		
		BENECAST_RAID_LIST = {};
				
	elseif ( event == 'ADDON_LOADED' and ( arg1 == 'Blizzard Raid UI' or arg1 == 'CT_RaidAssist' ) ) then
		if arg1 == 'CT_RaidAssist' then
			--Hook into CT_RA's update function
			if BC_Old_CTRA_UpdateFunction == nil and CT_RA_UpdateRaidGroup ~= nil then
				BC_Old_CTRA_UpdateFunction = CT_RA_UpdateRaidGroup;
				CT_RA_UpdateRaidGroup = BC_UpdateRaidGroup;
			end
		end
			
		BC_UpdateRaid = true;
	
	elseif ( event == 'ADDON_LOADED' ) then
		-- Load up the variables for the snapping if empty
		if BeneCastSetup.currentid == nil then
			BeneCastSetup.currentid = 1;
		end
		if BeneCastProfiles[BeneCastSetup.currentid] == nil then
			BeneCastProfiles[BeneCastSetup.currentid] = { SnapTo = 'BeneCast' };
		end
		if BeneCastProfiles[BeneCastSetup.currentid].SnapTo ~= 'CUSTOM' then
			BeneCast_SnapPanels(arg1);
		end
		
	elseif ( event == 'VARIABLES_LOADED' ) then
		if BeneCastSetup.currentid == nil then
			BeneCastSetup.currentid = 1;
		end
		if BeneCastProfiles[BeneCastSetup.currentid] == nil then
			BeneCastProfiles[BeneCastSetup.currentid] = {};
		end
		BeneCast_SnapPanels(BeneCastProfiles[BeneCastSetup.currentid].SnapTo);
		-- Wait with the snapping till everything is loaded
		BC_AttachPartyFrames = true;
	
	-- If SPELLCAST_FAILED then a BeneCast spell may have failed
	elseif ( event == 'SPELLCAST_FAILED' ) then
		-- A BeneCast spell failed if a BC_spellstarted and BC_sentmessage
		if ( BC_spellstarted and BC_sentmessage ) then
			BeneCast_PartyNotifyFail('SPELLCAST_FAILED');
		end
		-- Clear out everything used for party notification just in case
		BC_spellstarted = nil;
		BC_spellstopped = nil;
		BC_spelltarget = nil;
		BC_spellbeingcast = nil;
		BC_spellismax = nil;
		BC_sentmessage = nil;
		
	-- If SPELLCAST_INTERRUPTED then a BeneCast spell may have been interrupted
	elseif ( event == 'SPELLCAST_INTERRUPTED' ) then
		-- A BeneCast spell was interrupted by the player if BC_spellstopped and BC_sentmessage
		if ( BC_spellstopped and BC_sentmessage ) then
			BeneCast_PartyNotifyFail('SPELLCAST_CANCELED');
		end
		-- A BeneCast spell was interrupted by something else if BC_spellstarted and BC_sentmessage
		if ( BC_spellstarted and BC_sentmessage ) then
			BeneCast_PartyNotifyFail('SPELLCAST_INTERRUPTED');
		end
		-- Clear out everything just in case
		BC_spellstarted = nil;
		BC_spellstopped = nil;
		BC_spelltarget = nil;
		BC_spellbeingcast = nil;
		BC_spellismax = nil;
		BC_sentmessage = nil;
		
	elseif ( event == 'SPELLCAST_STOP' or event == 'UNIT_SPELLCAST_CHANNEL_STOP' ) then
		-- If BC_spellbeingcast and notification is on then this is an instant spell and you should do notification
		if ( BC_spellbeingcast ) then
			-- BC_sentmessage will be set to true of BeneCast_PartyNotify sends a message
			BC_sentmessage = BeneCast_PartyNotify(BC_spelltarget, BC_spellbeingcast, BC_spellismax, 0);
		-- Otherwise set _setmessage to false
		else
			BC_sentmessage = nil
		end
		-- If BC_spellstarted, then a BeneCast spell was started and you should set BC_sentmessage and BC_spellstopped to true to indicate that the spell was interrupted by the user
		if ( BC_spellstarted ) then
			BC_sentmessage = true;
			BC_spellstopped = true;
			BC_spellstarted = nil;
		else
			-- Clear this, just in case
			BC_spellstopped = nil;
		end
		-- Clear these out just in case
		BC_spelltarget = nil;
		BC_spellbeingcast = nil;
		BC_spellismax = nil;
		
	elseif ( event == 'SPELLCAST_START' or event == 'UNIT_SPELLCAST_CHANNEL_START' ) then
		-- If the spell being cast exists in BeneCast_SpellTypes, notification is on, and there is a BC_spellbeingcast then do notification
		if ( BeneCast_SpellTypes[BC_class][arg1] and BC_spellbeingcast ) then
			-- Set BC_sentmessage to true of BeneCast_PartyNotify sent a message
			BC_sentmessage = BeneCast_PartyNotify(BC_spelltarget, BC_spellbeingcast, BC_spellismax, arg2);
			-- Clear out BC_spelltarget, BC_spellbeingcast, and BC_spellismax so it isn't used by SPELLCAST_STOP
			BC_spelltarget = nil;
			BC_spellbeingcast = nil;
			BC_spellismax = nil;
			-- If you sent a message then BC_spellstarted to indicate failures/interruptions should send notifications
			if ( BC_sentmessage ) then
				BC_spellstarted = true;
			end
		else
			-- Clear out everything just in case
			BC_spellstarted = nil;
			BC_spellstopped = nil;
			BC_spelltarget = nil;
			BC_spellbeingcast = nil;
			BC_spellismax = nil;
			BC_sentmessage = nil;
		end
	
	-- We only want spell changes after entering the world
	elseif ( event == 'SPELLS_CHANGED' and arg1 == nil) then -- don't fire when user only browses spellbook
		
		-- Reset most of the tables for reloading of spells and reload the spell data
		BeneCast_ResetData();
		for i in pairs (BC_targets) do
			BeneCast_UpdateButtons(i);
		end
	
	-- Don't bother with UNIT_AURA if the player can attack the unit
	elseif ( ( event == 'UNIT_AURA' or event == 'UNIT_AURASTATE' ) and not UnitCanAttack('player', arg1) ) then
		if BC_AttachPartyFrames then
			BeneCast_AttachPartyPanelsFromVar();
			BC_AttachPartyFrames = nil;
		end
		-- Redraw the buttons for the UnitID in arg1
		if BeneCastSetup.Debug then
			local msg = arg1;
			if not msg then
				msg = 'nil';
			end
			PutDebugMsg(event .. ' for ' .. msg);
		end
		BeneCast_UpdateButtonsForUnit(arg1);
		
	elseif ( event == 'PLAYER_AURAS_CHANGED' ) then
		if BC_AttachPartyFrames then
			BeneCast_AttachPartyPanelsFromVar();
			BC_AttachPartyFrames = nil;
		end
		BeneCast_UpdateButtonsForUnit('player');
	
	elseif ( event == 'PARTY_MEMBERS_CHANGED' ) then
		-- Redraw the buttons for all partymembers (and partypets)
		for i = 1,4 do
			BC_UnitUpdater['party' .. i] = GetTime() + 1;
			BC_UnitUpdater['partypet' .. i] = GetTime() + 1;
		end
	
	-- Fired when the pet of a unit changes
	elseif ( event == 'UNIT_PET' ) then
		if arg1 == 'player' then
			BC_UnitUpdater['pet'] = GetTime() + 1;
		elseif string.find(arg1,'party') then
			local num;
			_, _, num = string.find(arg1,'%d+');
			if num then
				BC_UnitUpdater['partypet' .. num] = GetTime() + 1;
			end
		elseif string.find(arg1,'raid') then
			if BeneCastSetup.currentraidid then
				if BeneCastProfiles[BeneCastSetup.currentraidid] then
					if BeneCastProfiles[BeneCastSetup.currentraidid].ShowRaidPets then
						local num;
						_, _, num = string.find(arg1,'%d+');
						if num then
							BC_UnitUpdater['raidpet' .. num] = GetTime() + 1;
						end
					end
				end
			end
		end
		
	elseif ( event == 'PLAYER_TARGET_CHANGED' ) then
		PutDebugMsg('PLAYER_TARGET_CHANGED');
		BeneCast_AttachPartyPanelsFromVar();
		BC_AttachPartyFrames = nil;
		-- Redraw the buttons for the target
		BeneCast_UpdateButtonsForUnit('target');
		BC_UnitUpdater['target'] = GetTime() + 1;
		-- Redraw the buttons for the target of target
		BeneCast_UpdateButtonsForUnit('targettarget');
		BC_UnitUpdater['targettarget'] = GetTime() + 1;

	elseif ( event == 'PLAYER_FOCUS_CHANGED' ) then
		-- Redraw the buttons for the target
		BeneCast_AttachPanelsFromVar(93);
		BeneCast_UpdateButtonsForUnit('focus');
		BC_UnitUpdater['focus'] = GetTime() + 1;
	
	elseif ( event == 'UNIT_TARGET' and unit == 'target' ) then
		PutDebugMsg('UNIT_TARGET');
		-- Redraw the buttons for the target of target
		BeneCast_AttachPanelsFromVar(92);
		BeneCast_UpdateButtonsForUnit('targettarget');
		BC_UnitUpdater['targettarget'] = GetTime() + 1;
		
	elseif ( event == 'RAID_ROSTER_UPDATE' ) then
		PutDebugMsg('RAID_ROSTER_UPDATE event received');
		if not BeneCastSetup.currentraidid then
			BeneCastSetup.currentraidid = 1
		end
		local pf = BeneCastSetup.currentraidid;
		if BeneCastProfiles[pf] == nil then
			BeneCastProfiles[pf] = {};
		end
		if BeneCastProfiles[pf].RaidSnapTo == 'STANDARDBOTTOM' or BeneCastProfiles[pf].RaidSnapTo == 'STANDARDRIGHT' then
			BC_UpdateRaid = true;
		end
				
	elseif ( event == 'PLAYER_ALIVE' or event == 'PLAYER_DEAD' or event == 'PLAYER_UNGHOST' or event == 'UPDATE_SHAPESHIFT_FORMS' ) then
		-- These changes shouldn't affect who can be shown...
		-- so we only want to update those that we've been showing before. 
		for i in pairs (BC_targets) do
			BeneCast_UpdateButtons(i);
		end
		
	elseif ( event == 'UNIT_HEALTH' ) then
		-- Redraw a member's panel if they're dead
		if ( UnitHealth(arg1) <= 0 and UnitIsConnected(arg1) ) then
			BeneCast_UpdateButtonsForUnit(arg1);
		end
				
	elseif ( event == 'PLAYER_ENTERING_WORLD' ) then
		BeneCast_AttachPartyPanelsFromVar();
		this:RegisterEvent('ACTIONBAR_SLOT_CHANGED');
		this:RegisterEvent('SPELLS_CHANGED');
		this:RegisterEvent('SPELLCAST_FAILED');
		this:RegisterEvent('SPELLCAST_INTERRUPTED');
		this:RegisterEvent('SPELLCAST_STOP');
		this:RegisterEvent('SPELLCAST_START');
		this:RegisterEvent('PARTY_MEMBERS_CHANGED');
		this:RegisterEvent('PLAYER_TARGET_CHANGED');
		this:RegisterEvent('UNIT_AURA');
		this:RegisterEvent('PLAYER_AURAS_CHANGED');
		this:RegisterEvent('UNIT_AURASTATE');
		this:RegisterEvent('UPDATE_SHAPESHIFT_FORMS');
		this:RegisterEvent('PLAYER_ALIVE');
		this:RegisterEvent('PLAYER_DEAD');
		this:RegisterEvent('PLAYER_UNGHOST');
		this:RegisterEvent('RAID_ROSTER_UPDATE');
		this:RegisterEvent('UNIT_HEALTH');
		this:RegisterEvent('UNIT_PET');
	
	elseif ( event == 'PLAYER_LEAVING_WORLD' ) then
		this:UnregisterEvent('ACTIONBAR_SLOT_CHANGED');
		this:UnregisterEvent('SPELLS_CHANGED');
		this:UnregisterEvent('SPELLCAST_FAILED');
		this:UnregisterEvent('SPELLCAST_INTERRUPTED');
		this:UnregisterEvent('SPELLCAST_STOP');
		this:UnregisterEvent('SPELLCAST_START');
		this:UnregisterEvent('PARTY_MEMBERS_CHANGED');
		this:UnregisterEvent('PLAYER_TARGET_CHANGED');
		this:UnregisterEvent('UNIT_AURA');
		this:UnregisterEvent('PLAYER_AURAS_CHANGED');
		this:UnregisterEvent('UNIT_AURASTATE');
		this:UnregisterEvent('UPDATE_SHAPESHIFT_FORMS');
		this:UnregisterEvent('PLAYER_ALIVE');
		this:UnregisterEvent('PLAYER_DEAD');
		this:UnregisterEvent('PLAYER_UNGHOST');
		this:UnregisterEvent('RAID_ROSTER_UPDATE');
		this:UnregisterEvent('UNIT_HEALTH');
		this:UnregisterEvent('UNIT_PET');
	
	elseif ( event == 'PLAYER_REGEN_DISABLED' ) then
		-- Player entering combat
		BC_ConsiderInCombat = true;
		BeneCast_UpdateCurrentProfile();
		BeneCast_LoadOptions();
	
	elseif ( event == 'PLAYER_REGEN_ENABLED' ) then
		-- Player exiting combat
		BC_ConsiderInCombat = false;
		BeneCast_UpdateCurrentProfile();
		BeneCast_LoadOptions();
	
	end

end

function BeneCastPanelManager_OnUpdate()

	local hasMainHandEnchant, mainHandExpiration, mainHandCharges, hasOffHandEnchant, offHandExpiration, offHandCharges;
	
	hasMainHandEnchant, mainHandExpiration, mainHandCharges, hasOffHandEnchant, offHandExpiration, offHandCharges = GetWeaponEnchantInfo();
	
	if ( hasMainHandEnchant ) then
		local slotId, textureName;
	
		slotId, textureName = GetInventorySlotInfo('MainHandSlot');
	
		BeneCast_Tooltip:ClearLines();
		BeneCast_Tooltip:SetInventoryItem('player', slotId);
		
		for i = 1, BeneCast_Tooltip:NumLines() do
			if getglobal('BeneCast_TooltipTextLeft' .. i):GetText() then
				local line = getglobal('BeneCast_TooltipTextLeft' .. i):GetText();
				if ( BC_weaponenchant ~= BENECAST_STRINGS.WEAPONENCHANT_ROCKBITER and string.find(line, BENECAST_STRINGS.WEAPONENCHANT_ROCKBITER) ) then
					BC_weaponenchant = BENECAST_STRINGS.WEAPONENCHANT_ROCKBITER;
					BeneCast_UpdateButtons(1);
					do break end; --WINTROW.6, break for efficiency
				end
				if ( BC_weaponenchant ~= BENECAST_STRINGS.WEAPONENCHANT_FLAMETONGUE and string.find(line, BENECAST_STRINGS.WEAPONENCHANT_FLAMETONGUE) ) then
					BC_weaponenchant = BENECAST_STRINGS.WEAPONENCHANT_FLAMETONGUE;
					BeneCast_UpdateButtons(1);
					do break end; --WINTROW.6, break for efficiency
				end
				if ( BC_weaponenchant ~= BENECAST_STRINGS.WEAPONENCHANT_FROSTBRAND and string.find(line, BENECAST_STRINGS.WEAPONENCHANT_FROSTBRAND) ) then
					BC_weaponenchant = BENECAST_STRINGS.WEAPONENCHANT_FROSTBRAND;
					BeneCast_UpdateButtons(1);
					do break end; --WINTROW.6, break for efficiency
				end
				if ( BC_weaponenchant ~= BENECAST_STRINGS.WEAPONENCHANT_WINDFURY and string.find(line, BENECAST_STRINGS.WEAPONENCHANT_WINDFURY) ) then
					BC_weaponenchant = BENECAST_STRINGS.WEAPONENCHANT_WINDFURY;
					BeneCast_UpdateButtons(1);
					do break end; --WINTROW.6, break for efficiency
				end
				if ( BC_weaponenchant ~= BENECAST_STRINGS.WEAPONENCHANT_FEEDBACK and string.find(line, BENECAST_STRINGS.WEAPONENCHANT_FEEDBACK) ) then
					BC_weaponenchant = BENECAST_STRINGS.WEAPONENCHANT_FEEDBACK;
					BeneCast_UpdateButtons(1);
					do break end; --WINTROW.6, break for efficiency
				end
			end
		end
	else
		if ( BC_weaponenchant ) then
			BC_weaponenchant = nil;
			BeneCast_UpdateButtons(1);
		end
		BC_weaponenchant = nil;
	end

	-- Update all button-flashing at once
	local buffAlphaValue = nil;
	for i,buttonname in pairs (BC_buttons_that_flash) do
		local button = getglobal(buttonname);
		if button then
			if buffAlphaValue == nil then --this way it only has to be calculated once
				if ( BuffFrameFlashState == 1 ) then
					buffAlphaValue = (BUFF_FLASH_TIME_ON - BuffFrameFlashTime) / BUFF_FLASH_TIME_ON;
					buffAlphaValue = buffAlphaValue * (1 - BUFF_MIN_ALPHA) + BUFF_MIN_ALPHA;
				else
					buffAlphaValue = BuffFrameFlashTime / BUFF_FLASH_TIME_ON;
					buffAlphaValue = (buffAlphaValue * (1 - BUFF_MIN_ALPHA)) + BUFF_MIN_ALPHA;
					--this:SetAlpha(BuffFrameFlashTime / BUFF_FLASH_TIME_ON);
				end
			end
			button:SetAlpha(buffAlphaValue);
		end
	end
	
	if not BeneCastSetup.currentraidid then
		BeneCastSetup.currentraidid = 1;
	end
	local pf = BeneCastSetup.currentraidid;
	if BeneCastProfiles[pf] == nil then
		BeneCastProfiles[pf] = {};
	end
	if BeneCastProfiles[pf].ShowRaid then
		local docheck = false;
		if BC_UpdateRaid then
			if BC_Last_UpdateRaid_Done == nil then
				docheck = true;
			elseif (GetTime() - BC_Last_UpdateRaid_Done) >= 5 then
				docheck = true;
			end
		end
		if docheck then
			BC_Last_UpdateRaid_Done = GetTime();
			BC_UpdateRaid = nil;
			BeneCast_UpdateRaid();
		end
	end
		
	-- Check for buttons if they are within range
	if not BeneCastSetup.currentid then
		BeneCastSetup.currentid = 1;
	end
	local pf = BeneCastSetup.currentid;
	if BeneCastProfiles[pf] == nil then
		BeneCastProfiles[pf] = {};
	end
	if BeneCastProfiles[pf].CheckRange then
		local docheck = false;
		if ( BC_LastRangeCheck == nil ) or
		   ( (GetTime() - BC_LastRangeCheck) > 1 ) then -- Every 1 second
	   		docheck = true;
		end
		if docheck then
			BC_LastRangeCheck = GetTime();
			BeneCast_CheckRange();
		end
	end
	
	-- Check if any panels need updating
	local todelete = {};
	for unit, updatetime in pairs(BC_UnitUpdater) do
		if updatetime and (GetTime() > updatetime) then
			-- can't remove items while traversing the table
			-- it would mess things up
			table.insert(todelete,unit); 
			BeneCast_UpdateButtonsForUnit(unit);
		end
	end
	for k, unit in pairs(todelete) do
		BC_UnitUpdater[unit] = nil;
	end

end

function BeneCast_CheckRange()

	local inrange;
	for id = 2, table.getn(BC_targets) do
		local unit = BeneCast_GetUnitIDFromPanelID(id);
		if getglobal('BeneCastPanel' .. id):IsVisible() and UnitExists(unit) then
			for i = 1, BENECAST_MAX_NUMBER_OF_BUTTONS do
				local buttonname = 'BeneCastPanel' .. id .. 'Button' .. i;
				local button = getglobal(buttonname);
				if button:IsVisible() then
					local spellid = button.spellid;
					local range = getglobal(buttonname .. 'Range');
					range:Hide();
					if spellid then
						inrange = IsSpellInRange(spellid,unit);
						if (inrange == nil) or (inrange == 0) then
							range:Show();
							if unit == 'target' and BeneCastSetup.Debug then
								local spellname = GetSpellName(spellid, SpellBookFrame.bookType);
								PutDebugMsg('Button ' .. i .. ', spell ' .. spellname .. ' out of range');
							end
						end
					end
				end
			end
		end
	end

end

-- *****************************************************************************
-- BeneCastResetButton functions
-- *****************************************************************************

function BeneCast_ResetTooltip()

	GameTooltip:SetOwner(this, 'ANCHOR_RIGHT');
	GameTooltip:SetText(BENECAST_STRINGS['TEXT_RESET_BUTTON_TOOLTIP'], nil, nil, nil, nil, 1);

end

function BeneCast_ResetProfiles()

	--Reset the BeneCastProfiles-table
	if not BeneCastSetup then
		BeneCastSetup = {};
	end
	BeneCastProfiles = {};
	BeneCastProfiles[1] = { profilename = '<default>' };
	BeneCastSetup.currentid = 1;
	BeneCastSetup.currentraidid = 1;
	BeneCastSetup.currentoptionid = 1;
	BeneCastSetup.PrefferedInCombatProfile = 1;
	BeneCastSetup.PrefferedOutOfCombatProfile = 1;
	BeneCastSetup.Version = BENECAST_VERSION;
	
	BeneCast_ResetConfig(1);
end

function BeneCast.ResetConfig(id)

	-- Reset the specified Profile table
	local curid = BeneCastSetup.currentoptionid;
	if id == nil then
		id = curid
	end
	if BeneCastProfiles[id] == nil then
		BeneCastProfiles[id] = {};
	end
	
	local profname = BeneCastProfiles[id].profilename;
	BeneCastProfiles[id] = {};
	if profname == nil then
		profname = 'profile' .. id;
	end 
	BeneCastProfiles[id].profilename = profname;
	
	BeneCastProfiles[id][1] = {};
	BeneCastProfiles[id][2] = {};
	BeneCastProfiles[id][3] = {};
	BeneCastProfiles[id][4] = {};
	BeneCastProfiles[id][5] = {};
	BeneCastProfiles[id][6] = {};
	BeneCastProfiles[id][7] = {};
	BeneCastProfiles[id][8] = {};
	BeneCastProfiles[id][9] = {};
	BeneCastProfiles[id][10] = {};

	for i = 1,10 do
		for type in pairs (BeneCast_SpellTypes_Sorting[BC_class]) do
			BeneCastProfiles[id][i][type] = false;
		end
	end
	
	BeneCastProfiles[id]['ShowAllButtons'] = false;
	BeneCastProfiles[id]['OverlapRaidBars'] = false;
	BeneCastProfiles[id]['FlashAsFade'] = false;
	BeneCastProfiles[id]['HideMinimap'] = false;
	BeneCastProfiles[id]['ShowTooltips'] = true;
	BeneCastProfiles[id]['ShowTooltipsName'] = false;
	BeneCastProfiles[id]['Unlock'] = false;
	BeneCastProfiles[id]['Notification'] = false;
	BeneCastProfiles[id]['SelfNotify'] = true;
	BeneCastProfiles[id]['PartyNotify'] = false;
	BeneCastProfiles[id]['SayNotify'] = false;
	BeneCastProfiles[id]['RaidNotify'] = false;
	BeneCastProfiles[id]['TargetNotify'] = false;
	BeneCastProfiles[id]['UserNotify'] = false;
	BeneCastProfiles[id]['SelfCasts'] = true;
	BeneCastProfiles[id]['NotifyRank'] = true;
	BeneCastProfiles[id]['NotifyTime'] = true;
	BeneCastProfiles[id]['NotifyHeal'] = true;
	BeneCastProfiles[id]['NotifyMaxHeal'] = false;
	BeneCastProfiles[id]['NotifyCure'] = true;
	BeneCastProfiles[id]['NotifyBuff'] = true;
	BeneCastProfiles[id]['NotifyRes'] = true;
	BeneCastProfiles[id]['PlayerAsDefault'] = true;
	BeneCastProfiles[id]['ShowPets'] = false;
	BeneCastProfiles[id]['ShowRaid'] = false;
	BeneCastProfiles[id]['ShowRaidPets'] = false; 
	BeneCastProfiles[id]['CheckRange'] = false; 
	BeneCastProfiles[id]['ShowTargetTarget'] = true; 
	
	BeneCastProfiles[id]['ButtonSize'] = 36;
	BeneCastProfiles[id]['ButtonNumber'] = 5;
	
	BeneCastProfiles[id]['UserChannel'] = nil;
	
	BeneCastProfiles[id]['SnapTo'] = 'STANDARD';
	BeneCastProfiles[id]['SnapToBackup'] = 'STANDARD';
	BeneCastProfiles[id]['RaidSnapTo'] = 'STANDARDRIGHT';
	
	BeneCastProfiles[id]['Debug'] = false;
	
	if id == curid then
		BeneCast_ResetData();
		BeneCast_LoadOptions();
		-- Reset the SnapTo selections (Raid too, thnx Shimavak)
		UIDropDownMenu_SetSelectedValue(BeneCastSnapToMenu, BeneCastProfiles[id].SnapTo);
		UIDropDownMenu_SetSelectedValue(BeneCastRaidSnapToMenu, BeneCastProfiles[id].RaidSnapTo);
	end
	if id == BeneCastSetup.currentid then
		-- Place the panels in the correct place
		BeneCast_SnapPanels(BeneCastProfiles[id].SnapTo);
		BeneCast_AttachPartyPanelsFromVar();
	
		-- Redraw everyone's panels
		--players
		for i = 1, 5 do
			BeneCast_OrientButtons(i);
			BeneCast_UpdateButtons(i);
		end
		--pets
		for i = 46, 50 do
			BeneCast_OrientButtons(i);
			BeneCast_UpdateButtons(i);
		end
		--target, targettarget, focus
		for i = 91, 93 do
			BeneCast_OrientButtons(i);
			BeneCast_UpdateButtons(i);
		end
	end

end

-- *****************************************************************************
-- BeneCastUserChannel functions
-- *****************************************************************************

function BeneCastUserChannel_OnTextChange()

	if BeneCastSetup.currentoptionid == nil then
		BeneCastSetup.currentoptionid = 1;
	end
	local pf = BeneCastSetup.currentoptionid;
	if BeneCastProfiles[pf] == nil then
		BeneCastProfiles[pf] = {};
	end
	BeneCastProfiles[pf]['UserChannel'] = this:GetText();

end

-- *****************************************************************************
-- BeneCastEditBox functions
-- *****************************************************************************

function BeneCastEditBox_ApplyChange()
	
	if this.var1 == 'variable' then
		if this.var2 == 'BC_PanelNumber' then
			BC_PanelNumber = tonumber(this:GetText());
			BeneCastUpdateCustomEditBoxes();
		end
		
	elseif this.var1 == 'panelinfo' then
		local pf = BeneCast_ProfileForPanelID(BC_PanelNumber);
		local framename = 'BeneCastPanel' .. BC_PanelNumber;
		if BeneCastSetup.Debug then
			local msg = 'Updating BeneCastProfiles[' .. pf .. '][';
			if framename then
				msg = msg .. framename .. '][';
			else
				msg = msg .. 'nil][';
			end
			if this.var2 then
				msg = msg .. this.var2 .. ']';
			else
				msg = msg .. 'nil]';
			end
			PutDebugMsg(msg);
		end
		local boxvalue = this:GetText();
		if boxvalue ~= BeneCastProfiles[pf][framename][this.var2] then
			BeneCastProfiles[pf]['SnapTo'] = 'CUSTOM';
			UIDropDownMenu_SetSelectedValue(BeneCastSnapToMenu, 'CUSTOM');
			BeneCastProfiles[pf][framename][this.var2] = boxvalue;
			BeneCastProfiles[pf][framename]['Moved'] = false;
		end
		PutDebugMsg('Updating Panel ' .. BC_PanelNumber);
		BeneCast_AttachPanelsFromVar(tonumber(BC_PanelNumber));
		
	elseif this.var1 == 'profilename' then
		local newtext = this:GetText();
		if nextext == '' then
			newtext = nil;
		end
		if newtext ~= this.oldtext then
			local profileid;
			if this.oldtext then
				for k,v in pairs(BeneCastProfiles) do
					if v.profilename == this.oldtext then
						profileid = k;
						do break end;
					end
				end
			end
			local newid;
			if newtext then
				for k,v in pairs(BeneCastProfiles) do
					if v.profilename == newtext then
						newid = k;
						do break end;
					end
				end
			end
			if newid then
				PutMsg('There is already a profile with that name');
				this:SetText(this.oldtext);
				return;
			end
			-- Change name of existing profile
			if profileid and newtext then
				BeneCastProfiles[profileid].profilename = newtext;
			-- Create new profile
			elseif not profileid and newtext then
				newid = table.getn(BeneCastProfiles)+1;
				BeneCastProfiles[newid] = {};
				BeneCastProfiles[newid].profilename = newtext;
				BeneCast_ResetConfig(newid);
				
			-- Delete existing profile
			elseif profileid and not newtext then
				table.remove(BeneCastProfiles,profileid);
				if table.getn(BeneCastProfiles) == 0 then
					BeneCast_ResetProfiles();
				end
				for i, v in pairs(BeneCastProfiles) do
					newid = i;
					do break end;
				end
				if BeneCastSetup['PrefferedInCombatProfile'] == profileid then
					BeneCastSetup['PrefferedInCombatProfile'] = newid;
				end
				if BeneCastSetup['PrefferedOutOfCombatProfile'] == profileid then
					BeneCastSetup['PrefferedOutOfCombatProfile'] = newid;
				end
				if BeneCastSetup.currentid == profileid then
					BeneCastSetup.currentid = nil;
					BeneCast_UpdateCurrentProfile();
				end
				BeneCast_LoadOptions();
			end
			BeneCast_UpdateProfileCheckButtons();
			this.oldtext = newtext;
		end
	
	elseif this.var1 == 'currprofile' then
		if this:GetText() == nil then
			this:SetText('1');
		elseif this:GetText() == '' then
			this:SetText('1');
		end
		local newid = tonumber(this:GetText());
		if newid ~= BeneCastSetup.currentoptionid then
			BeneCastSetup.currentoptionid = newid;
			BeneCast_LoadOptions();
			BeneCast_LoadSpellConfig();
		end
		
	end
	
end

-- *****************************************************************************
-- BeneCastSnapToMenu functions
-- *****************************************************************************

function BeneCastSnapToMenu_Initialize()
	
	local info;
	-- Add SnapTo options
	if BeneCastSetup.currentoptionid == nil then
		BeneCastSetup.currentoptionid = 1;
	end
	local pf = BeneCastSetup.currentoptionid;
	if BeneCastProfiles[pf] == nil then
		BeneCastProfiles[pf] = {};
	end
	for snapto in pairs (BeneCast_SnapTo) do
		if IsAddOnLoaded(BeneCast_SnapTo[snapto].AddOn) or snapto == 'CUSTOM' then
			info = {};
			if ( BeneCast_SnapTo[snapto].AddOn == 'BeneCast' ) then
				info.text = BENECAST_STRINGS.TEXT_STANDARD;
			elseif ( BeneCast_SnapTo[snapto].AddOn == 'Perl_Target' ) then
				info.text = BENECAST_STRINGS.TEXT_PERLCLASSIC;
			elseif ( snapto == 'CUSTOM' ) then
				info.text = 'Custom';
			else
				info.text = BeneCast_SnapTo[snapto].AddOn;
			end
			
			info.value = snapto;
			if ( BeneCastProfiles[pf].SnapTo == snapto ) then
				info.checked = true;
			else
				info.checked = false;
			end
			info.func = BeneCastSnapToMenu_SetSnap;
			UIDropDownMenu_AddButton(info, UIDROPDOWNMENU_MENU_LEVEL);
		end
	end

end

function SetBeneCastSnapTo(snapto_option)
	local found;
	for snaptoval in pairs (BeneCast_SnapTo) do
		if BeneCast_SnapTo[snaptoval].AddOn == snapto_option then
			if BeneCastSetup.currentoptionid == nil then
				BeneCastSetup.currentoptionid = 1;
			end
			local pf = BeneCastSetup.currentoptionid;
			if BeneCastProfiles[pf] == nil then
				BeneCastProfiles[pf] = {};
			end
			found = true;
			BeneCastProfiles[pf].SnapTo = snaptoval;
			if snaptoval ~= 'CUSTOM' then
				BeneCastProfiles[pf].SnapToBackup = snaptoval;
			end
			BeneCast_SnapPanels(snapto_option);
			BeneCast_AttachPartyPanelsFromVar(); --WINTROW.6
			UIDropDownMenu_SetSelectedValue(BeneCastSnapToMenu, snaptoval);
			for i in pairs (BC_targets) do
				BeneCast_UpdateButtons(i);
			end
			do break end;
		end
	end
	if not found then
		PutDebugMsg(snapto_option .. ' not found in predefined unit frames');
	end
end

function SetBeneCastSnapTo2(snapto_option)
	if ( not BeneCast_SnapTo[snapto_option] ) then
		PutDebugMsg(snapto_option .. ' not found in predefined unit frames');
		return;
	end
	if BeneCastSetup.currentoptionid == nil then
		BeneCastSetup.currentoptionid = 1;
	end
	local pf = BeneCastSetup.currentoptionid;
	if BeneCastProfiles[pf] == nil then
		BeneCastProfiles[pf] = {};
	end
	BeneCastProfiles[pf].SnapTo = snapto_option;
	if snapto_option ~= 'CUSTOM' then
		BeneCastProfiles[pf].SnapToBackup = snapto_option;
	end
	BeneCast_SnapPanels(BeneCast_SnapTo[snapto_option].AddOn);
	BeneCast_AttachPartyPanelsFromVar(); --WINTROW.6
	UIDropDownMenu_SetSelectedValue(BeneCastSnapToMenu, snapto_option);
	for i in pairs (BC_targets) do
		BeneCast_UpdateButtons(i);
	end
end

function PutMsg(msg)
	if ( DEFAULT_CHAT_FRAME ) then
		DEFAULT_CHAT_FRAME:AddMessage(msg);
	end
end

function BeneCastSnapToMenu_SetSnap()

	if ( not this.value ) then
		return;
	end
	local pf = BeneCastSetup.currentoptionid;
	if BeneCastProfiles[pf] == nil then
		BeneCastProfiles[pf] = {};
	end
	PutDebugMsg('Entered BeneCastSnapToMenu_SetSnap'); --WINTROW.6
	BeneCastProfiles[pf].SnapTo = this.value;
	if this.value ~= 'CUSTOM' then
		BeneCastProfiles[pf].SnapToBackup = this.value;
	end
	PutDebugMsg('SnapTo = ' .. this.value);
	BeneCast_SnapPanels(BeneCast_SnapTo[this.value].AddOn,true);
	BeneCast_AttachPartyPanelsFromVar();
	UIDropDownMenu_SetSelectedValue(BeneCastSnapToMenu, this.value);
	for i in pairs (BC_targets) do
		BeneCast_UpdateButtons(i);
	end
	PutDebugMsg('Updated buttons');

end

function BeneCastSnapToMenu_Load()

	if BeneCastSetup.currentoptionid == nil then
		BeneCastSetup.currentoptionid = 1;
	end
	local pf = BeneCastSetup.currentoptionid;
	if BeneCastProfiles[pf] == nil then
		BeneCastProfiles[pf] = {};
	end
	UIDropDownMenu_Initialize(this, BeneCastSnapToMenu_Initialize);
	UIDropDownMenu_SetSelectedValue(this, BeneCastProfiles[pf].SnapTo);
	getglobal('BeneCastSnapToMenuLabel'):SetText(BENECAST_STRINGS.TEXT_SNAP_TO);
	BeneCastSnapToMenu.tooltip = BENECAST_STRINGS.TEXT_SNAP_TO_TOOLTIP;
	UIDropDownMenu_SetWidth(110, BeneCastSnapToMenu);

end

-- *****************************************************************************
-- BeneCastRaidSnapToMenu functions
-- *****************************************************************************

function BeneCastRaidSnapToMenu_Initialize()
	
	local info;
	-- Add SnapTo options
	if BeneCastSetup.currentoptionid == nil then
		BeneCastSetup.currentoptionid = 1;
	end
	local pf = BeneCastSetup.currentoptionid;
	if BeneCastProfiles[pf] == nil then
		BeneCastProfiles[pf] = {};
	end
	for snapto in pairs (BeneCast_RaidSnapTo) do
		if ( IsAddOnLoaded(BeneCast_RaidSnapTo[snapto].AddOn) ) then
			info = {};
			if ( BeneCast_RaidSnapTo[snapto].AddOn == 'BeneCast' ) then
				info.text = BENECAST_STRINGS.TEXT_STANDARD;
			else
				info.text = BeneCast_RaidSnapTo[snapto].AddOn;
			end
			-- Determine the position for BeneCast panels to snap to
			if ( string.find(snapto, 'BOTTOM') ) then
				info.text = info.text .. BENECAST_STRINGS.TEXT_BOTTOM;
			elseif ( string.find(snapto, 'LEFT') ) then
				info.text = info.text .. BENECAST_STRINGS.TEXT_LEFT;
			elseif ( string.find(snapto, 'RIGHT') ) then
				info.text = info.text .. BENECAST_STRINGS.TEXT_RIGHT;
			elseif ( string.find(snapto, 'TOP') ) then
				info.text = info.text .. BENECAST_STRINGS.TEXT_TOP;
			end
			
			info.value = snapto;
			if ( BeneCastProfiles[pf].RaidSnapTo == snapto ) then
				info.checked = true;
			else
				info.checked = false;
			end
			info.func = BeneCastRaidSnapToMenu_SetSnap;
			UIDropDownMenu_AddButton(info, UIDROPDOWNMENU_MENU_LEVEL);
		end
	end

end

function BeneCastRaidSnapToMenu_SetSnap()

	if BeneCastSetup.currentoptionid == nil then
		BeneCastSetup.currentoptionid = 1;
	end
	local pf = BeneCastSetup.currentoptionid;
	if BeneCastProfiles[pf] == nil then
		BeneCastProfiles[pf] = {};
	end
	BeneCastProfiles[pf].RaidSnapTo = this.value;
	UIDropDownMenu_SetSelectedValue(BeneCastRaidSnapToMenu, this.value);
	BC_UpdateRaid = true;
	
end

function BeneCastRaidSnapToMenu_Load()

	if BeneCastSetup.currentoptionid == nil then
		BeneCastSetup.currentoptionid = 1;
	end
	local pf = BeneCastSetup.currentoptionid;
	if BeneCastProfiles[pf] == nil then
		BeneCastProfiles[pf] = {};
	end
	UIDropDownMenu_Initialize(this, BeneCastRaidSnapToMenu_Initialize);
	UIDropDownMenu_SetSelectedValue(this, BeneCastProfiles[pf].RaidSnapTo);
	getglobal('BeneCastRaidSnapToMenuLabel'):SetText(BENECAST_STRINGS.TEXT_RAID_SNAP_TO);
	BeneCastRaidSnapToMenu.tooltip = BENECAST_STRINGS.TEXT_RAID_SNAP_TO_TOOLTIP;

end

-- *****************************************************************************
-- BeneCastNotificationDropDown functions
-- *****************************************************************************

function BeneCastNotificationDropDown_Initialize()
	
	if BeneCastSetup.currentoptionid == nil then
		BeneCastSetup.currentoptionid = 1;
	end
	local pf = BeneCastSetup.currentoptionid;
	if BeneCastProfiles[pf] == nil then
		BeneCastProfiles[pf] = {};
	end

	local info, buttonIndex;

	-- Add the menu title which will be the default selected item.
	info = {};
	info.text = BENECAST_STRINGS.TEXT_USER_CHANNEL;
	info.value = 'Tip';
	info.isTitle = true;
	info.checked = false;
	info.owner = this;

	UIDropDownMenu_AddButton(info, UIDROPDOWNMENU_MENU_LEVEL);

	-- Finds the index of the associated dropdown
	buttonIndex = this:GetID();
	if buttonIndex == 0 then
		buttonIndex = this:GetParent():GetID();
	end
	
	-- Finds the config of the current dropdown box
	local values = BeneCastOptionDropDowns[buttonIndex];
	if values == nil then
		PutDebugMsg('no BeneCastOptionDropDowns for id ' .. buttonIndex .. ' object ' .. this:GetName());
		return;
	end

	-- Add Available notification targets
	info = {};
	info.arg1 = {};
	for textReference, channel in pairs (BeneCastNotificationTargets) do
		info.text = BENECAST_STRINGS[textReference];
		info.value = channel;
		info.arg1.chan = channel;
		info.arg1.cvar = values.cvar;
		info.arg1.dropdown = this;
		info.arg1.index = buttonIndex;
		info.func = BeneCastNotificationDropDown_OnClick;
		info.tooltipTitle = BENECAST_STRINGS[textReference];
		info.tooltipText = BENECAST_STRINGS[textReference .. '_TOOLTIP'];
		info.checked = (BeneCastProfiles[pf][values.cvar] == channel);
		info.owner = this;
		
		UIDropDownMenu_AddButton(info, UIDROPDOWNMENU_MENU_LEVEL);
	end

end

function BeneCastNotificationDropDown_OnClick(arg1)

	if BeneCastSetup.currentoptionid == nil then
		BeneCastSetup.currentoptionid = 1;
	end
	local pf = BeneCastSetup.currentoptionid;
	if BeneCastProfiles[pf] == nil then
		BeneCastProfiles[pf] = {};
	end
	-- Finds the index of the associated dropdown
	PutDebugMsg('Clicked on ' .. arg1.dropdown:GetName() .. ' id=' .. arg1.index);
	if BeneCastOptionDropDowns[buttonIndex] then
		if BeneCastProfiles[pf].CascadeChannels then
			for i,values in pairs (BeneCastOptionDropDowns) do
				BeneCastProfiles[pf][values.cvar] = arg1.chan;
				UIDropDownMenu_SetSelectedValue(getglobal('BeneCastOptionDropDown' .. i),arg1.chan);
			end
		else
			BeneCastProfiles[pf][arg1.cvar] = arg1.chan;
			UIDropDownMenu_SetSelectedValue(arg1.dropdown, arg1.chan);
		end
	end
	
end

function BeneCastNotificationDropDown_Load()

	if BeneCastSetup.currentoptionid == nil then
		BeneCastSetup.currentoptionid = 1;
	end
	local pf = BeneCastSetup.currentoptionid;
	if BeneCastProfiles[pf] == nil then
		BeneCastProfiles[pf] = {};
	end
	-- Finds the index of the associated dropdown
	local buttonIndex = this:GetID();
	if buttonIndex == 0 then
		buttonIndex = this:GetParent():GetID();
	end
	local cvar = BeneCastOptionDropDowns[buttonIndex].cvar;
	UIDropDownMenu_Initialize(this, BeneCastNotificationDropDown_Initialize);
	if not BeneCastProfiles[pf][cvar] then
		BeneCastProfiles[pf][cvar] = 'SELF';
	end
	UIDropDownMenu_SetSelectedValue(this,BeneCastProfiles[pf][cvar]);
end

-- *****************************************************************************
-- BeneCastOptionFrame functions
-- *****************************************************************************

function BeneCastOptionFrame_OnLoad()
	
	-- Binding labels
	BINDING_HEADER_BENECAST = 'BeneCast';
	BINDING_NAME_BENECAST_OPTIONS = BENECAST_STRINGS.TEXT_BENECAST_OPTIONS;
	BINDING_NAME_BENECAST_TOGGLE_BUFFS = BENECAST_STRINGS.TEXT_BENECAST_TOGGLE_BUFFS;
	BINDING_NAME_BENECAST_BUTTON1 = BENECAST_STRINGS.TEXT_BENECAST_BUTTON1;
	BINDING_NAME_BENECAST_BUTTON2 = BENECAST_STRINGS.TEXT_BENECAST_BUTTON2;
	BINDING_NAME_BENECAST_BUTTON3 = BENECAST_STRINGS.TEXT_BENECAST_BUTTON3;
	BINDING_NAME_BENECAST_BUTTON4 = BENECAST_STRINGS.TEXT_BENECAST_BUTTON4;
	BINDING_NAME_BENECAST_BUTTON5 = BENECAST_STRINGS.TEXT_BENECAST_BUTTON5;
	BINDING_NAME_BENECAST_BUTTON6 = BENECAST_STRINGS.TEXT_BENECAST_BUTTON6;
	BINDING_NAME_BENECAST_BUTTON7 = BENECAST_STRINGS.TEXT_BENECAST_BUTTON7;
	BINDING_NAME_BENECAST_BUTTON8 = BENECAST_STRINGS.TEXT_BENECAST_BUTTON8;
	BINDING_NAME_BENECAST_BUTTON9 = BENECAST_STRINGS.TEXT_BENECAST_BUTTON9;
	BINDING_NAME_BENECAST_BUTTON10 = BENECAST_STRINGS.TEXT_BENECAST_BUTTON10;
	SLASH_BENECAST1 = '/benecast';
	SLASH_BENECAST2 = '/bc';
	
	-- Setup the BeneCast Slash Command Handler
	SlashCmdList['BENECAST'] = function(msg)
		BeneCast_SlashCommandHandler(msg);
	end
	
	-- Set up the title in the option frame
	getglobal('BeneCastTitleText'):SetText(BENECAST_STRINGS.TEXT_TITLE);

	-- Check Tab 1 and show the Player spell selection subframe
	getglobal('BeneCastPlayerFrame'):Show();
	getglobal('BeneCastOptionFrameTab1'):SetChecked(1);
	getglobal('BeneCastOptionFrameTab1'):Disable();
	
	-- Setup Tab Text and Graphics
	getglobal('BeneCastOptionFrameTab1NormalTexture'):SetTexture('Interface\\Icons\\Spell_Nature_WispSplode');
	getglobal('BeneCastOptionFrameTab1').tooltipText = UnitName('player');
	
	getglobal('BeneCastOptionFrameTab2NormalTexture'):SetTexture('Interface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes');
	getglobal('BeneCastOptionFrameTab2NormalTexture'):SetTexCoord(0.7421875, 0.98828125, 0, 0.25);
	getglobal('BeneCastOptionFrameTab2').tooltipText = BENECAST_STRINGS.CLASS_DRUID;
	
	getglobal('BeneCastOptionFrameTab3NormalTexture'):SetTexture('Interface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes');
	getglobal('BeneCastOptionFrameTab3NormalTexture'):SetTexCoord(0, 0.25, 0.25, 0.5);
	getglobal('BeneCastOptionFrameTab3').tooltipText = BENECAST_STRINGS.CLASS_HUNTER;
	
	getglobal('BeneCastOptionFrameTab4NormalTexture'):SetTexture('Interface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes');
	getglobal('BeneCastOptionFrameTab4NormalTexture'):SetTexCoord(0.25, 0.49609375, 0, 0.25);
	getglobal('BeneCastOptionFrameTab4').tooltipText = BENECAST_STRINGS.CLASS_MAGE;
	
	getglobal('BeneCastOptionFrameTab5NormalTexture'):SetTexture('Interface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes');
	getglobal('BeneCastOptionFrameTab5NormalTexture'):SetTexCoord(0.49609375, 0.7421875, 0.25, 0.5);
	getglobal('BeneCastOptionFrameTab5').tooltipText = BENECAST_STRINGS.CLASS_PRIEST;
	
	getglobal('BeneCastOptionFrameTab6NormalTexture'):SetTexture('Interface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes');
	getglobal('BeneCastOptionFrameTab6NormalTexture'):SetTexCoord(0.49609375, 0.7421875, 0, 0.25);
	getglobal('BeneCastOptionFrameTab6').tooltipText = BENECAST_STRINGS.CLASS_ROGUE;
	
	getglobal('BeneCastOptionFrameTab7NormalTexture'):SetTexture('Interface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes');
	getglobal('BeneCastOptionFrameTab7NormalTexture'):SetTexCoord(0.7421875, 0.98828125, 0.25, 0.5);
	getglobal('BeneCastOptionFrameTab7').tooltipText = BENECAST_STRINGS.CLASS_WARLOCK;
	
	getglobal('BeneCastOptionFrameTab8NormalTexture'):SetTexture('Interface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes');
	getglobal('BeneCastOptionFrameTab8NormalTexture'):SetTexCoord(0, 0.25, 0, 0.25);
	getglobal('BeneCastOptionFrameTab8').tooltipText = BENECAST_STRINGS.CLASS_WARRIOR;

	getglobal('BeneCastOptionFrameTab9NormalTexture'):SetTexture('Interface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes');
	getglobal('BeneCastOptionFrameTab9NormalTexture'):SetTexCoord(0.25, 0.49609375, 0.25, 0.5);
	getglobal('BeneCastOptionFrameTab9').tooltipText = BENECAST_STRINGS.CLASS_SHAMAN;
	
	getglobal('BeneCastOptionFrameTab10NormalTexture'):SetTexture('Interface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes');
	getglobal('BeneCastOptionFrameTab10NormalTexture'):SetTexCoord(0.0, 0.25, 0.5, 0.75);
	getglobal('BeneCastOptionFrameTab10').tooltipText = BENECAST_STRINGS.CLASS_PALADIN;
	
	getglobal('BeneCastOptionFrameTab11NormalTexture'):SetTexture('Interface\\Icons\\INV_Misc_Wrench_01');
	getglobal('BeneCastOptionFrameTab11').tooltipText = BENECAST_STRINGS.TEXT_SETUP;
	
	getglobal('BeneCastOptionFrameTab12NormalTexture'):SetTexture('Interface\\Icons\\INV_Letter_08');
	getglobal('BeneCastOptionFrameTab12').tooltipText = BENECAST_STRINGS.TEXT_NOTIFICATION;
	
	getglobal('BeneCastOptionFrameTab13NormalTexture'):SetTexture('Interface\\Icons\\INV_Misc_Head_Dragon_01');
	getglobal('BeneCastOptionFrameTab13').tooltipText = BENECAST_STRINGS.TEXT_RAID;

	getglobal('BeneCastOptionFrameTab14NormalTexture'):SetTexture('Interface\\Icons\\INV_Misc_Wrench_02');
	getglobal('BeneCastOptionFrameTab14').tooltipText = 'Unit Frames';

end

-- *****************************************************************************
-- BeneCastMiniMapButton Game Event Handlers
-- *****************************************************************************

-- Clicks on the BeneCast button do different things
function BeneCastOptionFrameToggle()

	-- If the mouseButton is the left button toggle BeneCastFrame's visibility
	-- if ( arg1 == 'LeftButton' ) then
		local frame = getglobal('BeneCastOptionFrame');
		if ( frame:IsVisible() ) then
			frame:Hide();
		else
			frame:Show();
		end
	--end
	
end

function BeneCastMiniMap_OnMouseDown()

	if BeneCastSetup.currentid == nil then
		BeneCastSetup.currentid = 1;
	end
	local pf = BeneCastSetup.currentid;
	if BeneCastProfiles[pf] == nil then
		BeneCastProfiles[pf] = {};
	end
	if ( ( BeneCastProfiles[pf].Unlock ) and ( arg1 == 'LeftButton' ) ) then
		this:StartMoving();
		this.isMoving = true;
	end

end

function BeneCastMiniMap_OnLoad()

	this:RegisterEvent('ADDON_LOADED');

end

function BeneCastMiniMap_OnEvent()

	if ( event == 'ADDON_LOADED' and arg1 == 'BeneCast' ) then
		if BeneCastSetup.currentid == nil then
			BeneCastSetup.currentid = 1;
		end
		local pf = BeneCastSetup.currentid;
		if BeneCastProfiles[pf] == nil then
			BeneCastProfiles[pf] = {};
		end
		if ( BeneCastProfiles[pf].HideMinimap ) then
			getglobal('BeneCastMinimapButton'):Hide();
		end
	end

end
